/*******************************************************************************
 * Copyright 2015 - 2019 fmCreation
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald diese von der
 * Europischen Kommission genehmigt wurden - Folgeversionen der EUPL
 * ("Lizenz"); Sie drfen dieses Werk ausschlielich gem dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 *   https://progs.frankmuenster.de/eupl_v1-2_de
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften gefordert oder in
 * schriftlicher Form vereinbart, wird die unter der Lizenz verbreitete Software
 * "so wie sie ist", OHNE JEGLICHE GEWHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrcklich oder stillschweigend - verbreitet. Die sprachspezifischen
 * Genehmigungen und Beschrnkungen unter der Lizenz sind dem Lizenztext zu
 * entnehmen.
 ******************************************************************************/

package de.frankmuenster.mahoe.banks.targobank.web; // NOPMD

import java.io.IOException;
import java.net.URI;
import java.net.URISyntaxException;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;

import org.apache.commons.lang.StringUtils;
import org.openqa.selenium.By;
import org.openqa.selenium.NoSuchElementException;
import org.openqa.selenium.WebDriver;
import org.openqa.selenium.WebElement;
import org.openqa.selenium.support.ui.Select;

import de.frankmuenster.mahoe.banks.targobank.TargoBank;
import de.frankmuenster.mahoe.banks.targobank.pdfextractor.TargoBankPdfExtract;
import de.frankmuenster.mahoe.pdfextractor.AbstractPdfExtractor;
import de.frankmuenster.mahoe.pdfextractor.InvalidCardNumber;
import de.frankmuenster.mahoe.pdfextractor.PdfBuchung;
import de.frankmuenster.mahoe.utils.I18N;
import de.frankmuenster.mahoe.utils.TransactionSaver;
import de.frankmuenster.mahoe.utils.Utils;
import de.frankmuenster.mahoe.utils.WebUtils;
import de.frankmuenster.mahoe.web.AbstractWebSite;
import de.frankmuenster.mahoe.web.BankAccountStatementPage;
import de.frankmuenster.mahoe.web.WebSiteException;
import de.frankmuenster.mahoe.web.WebSiteExceptionTypes;
import de.willuhn.jameica.hbci.rmi.Umsatz;
import de.willuhn.logging.Logger;

/**
 * An object representing the account statements web page of the TargoBank
 *
 * @author Frank Mnster
 * @since 2.3.0
 */
public class TargoBankAccountStatementsPage implements BankAccountStatementPage {

    /**
     * Service not available text
     */
    private static final String CHECK_SERVICE = "Der Dienst ist derzeit nicht verf"; //$NON-NLS-1$

    /**
     * Kennung fr eine Kreditkartenabrechnung
     */
    private static final String EBSCL1 = "EBSCL1"; //$NON-NLS-1$

    /**
     * Minimun cell size of a table row with account sheet PDFs
     */
    private static final int MIN_ROW_CELL_SIZE = 2;

    /**
     * The account info for the statements
     */
    private final String cardAccount;

    /**
     * This is the driver instance.
     */
    private final WebDriver driver;

    /**
     * Did we hit the last PDF based on a given date
     */
    private boolean lastPdfToScanHit;

    /**
     * Constructs a new statement page object with the first page loaded
     *
     * @param website
     * @param cardAccount
     * @throws WebSiteException
     */
    public TargoBankAccountStatementsPage(final WebDriver driver, final String cardAccount) throws WebSiteException {

        this.cardAccount = cardAccount;
        this.driver = driver;

        this.getAccountSheetsPage();

    }

    /*
     * (non-Javadoc)
     *
     * @see de.frankmuenster.mahoe.banks.targobank.web.BankAccountStatementPage#
     * accountLinksHasNextPage()
     */
    @Override
    public boolean accountLinksHasNextPage() {
        return !this.lastPdfToScanHit && this.hasNextPage();
    }

    /**
     * Read all links on the page till the given date and stores them as a
     * {@link WebElement} together with the date in a map.
     *
     * @param from
     * @return a map with all found links
     */
    public Map<Date, WebElement> getAccountLinks(final Date from) {

        // reset lastPdfToScanHit
        this.lastPdfToScanHit = false;

        final Map<Date, WebElement> links = new ConcurrentHashMap<>();

        final SimpleDateFormat sdf = new SimpleDateFormat("dd.MM.yyyy", Locale.GERMAN); //$NON-NLS-1$

        final WebElement tableList = this.driver.findElement(By.id("list")); //$NON-NLS-1$

        final WebElement body = tableList.findElement(By.xpath("table/tbody")); //$NON-NLS-1$
        final List<WebElement> rows = body.findElements(By.tagName(WebUtils.HTML_TAG_TR));

        for (final WebElement row : rows) {

            Logger.debug(row.getText());

            // check if all cells are present
            if (row.findElements(By.tagName(WebUtils.HTML_TAG_TD))
                    .size() < TargoBankAccountStatementsPage.MIN_ROW_CELL_SIZE) {
                // If not simply continue
                continue;
            }

            // get date
            final String pdfDateString = row.findElement(By.xpath("td[2]")).getText().trim() //$NON-NLS-1$
                    .replaceAll(AbstractWebSite.NON_BREAK_SPACE, StringUtils.EMPTY);
            Logger.debug(I18N.tr("TargoBankAccountStatementsPage.billingDate", pdfDateString)); //$NON-NLS-1$
            try {
                final Date pdfDate = sdf.parse(pdfDateString);
                if (pdfDate.before(from)) {
                    // We are before the given date to check, so we can stop
                    // further processing
                    Logger.debug(I18N.tr("TargoBankAccountStatementsPage.reachedLastPdf")); //$NON-NLS-1$
                    this.lastPdfToScanHit = true;
                    break;
                } else {
                    // getthe link to teh PDF
                    final WebElement link = row.findElement(By.xpath("td[3]//a")); //$NON-NLS-1$
                    Logger.debug("link: " + link.getText()); //$NON-NLS-1$

                    if (link != null && link.getText().toUpperCase(Locale.getDefault())
                            .contains(TargoBankAccountStatementsPage.EBSCL1)) {
                        // Add the link together with the date to the list
                        links.put(pdfDate, link);
                    }
                }
            } catch (final ParseException e1) {
                Logger.error(I18N.tr("TargoBankAccountStatementsPage.readPdfDateFailed", e1.getMessage()), e1); //$NON-NLS-1$

            } catch (final NoSuchElementException re) {
                Logger.error(I18N.tr("TargoBankAccountStatementsPage.createUmsatzError", re.getMessage()), re); //$NON-NLS-1$
            }
        } // end for weRows

        return links;
    }

    /*
     * (non-Javadoc)
     *
     * @see de.frankmuenster.mahoe.banks.targobank.web.BankAccountStatementPage#
     * getNextPage()
     */
    @Override
    public BankAccountStatementPage getNextPage() {

        // only if a next page exists
        if (hasNextPage()) {
            final WebElement nextBtn = this.driver.findElement(By.id("NEXT")); //$NON-NLS-1$
            nextBtn.click();
        }
        return this;

    }

    /*
     * (non-Javadoc)
     *
     * @see de.frankmuenster.mahoe.banks.targobank.web.BankAccountStatementPage#
     * getTransactions(java.util.Date)
     */
    @Override
    public List<Umsatz> getTransactions(final Date startDate) {

        final List<Umsatz> umsaetze = new ArrayList<>();
        // get all the links to the the PDFs from the given start date
        final Map<Date, WebElement> links = this.getAccountLinks(startDate);

        for (final WebElement link : links.values()) {
            final List<Umsatz> transactions = this.loadTransactionsFromPDF(link);
            umsaetze.addAll(transactions);
        }

        return umsaetze;
    }

    /*
     * (non-Javadoc)
     *
     * @see de.frankmuenster.mahoe.banks.targobank.web.BankAccountStatementPage#
     * hasNextPage()
     */
    @Override
    public boolean hasNextPage() {

        return this.driver.findElements(By.id("NEXT")).size() == 1; //$NON-NLS-1$

    }

    /**
     * Search the open document page for the current account and open the page
     * with the list of account sheets only from this card account.
     *
     * @throws WebSiteException
     *             in case the account could not be found
     */
    private void getAccountSheetsPage() throws WebSiteException {

        // ffnen der PDF Liste
        Logger.debug(I18N.tr("TargoBankAccountStatementsPage.openStatementPageFor", //$NON-NLS-1$
                Utils.maskCardNummer(this.cardAccount)));

        // open transaction sheet page
        this.driver.get(TargoBank.URL_ABBRECHNUNGEN);
        // check for service error

        final WebElement body = this.driver.findElement(By.tagName("body")); //$NON-NLS-1$

        if (body.getText().contains(TargoBankAccountStatementsPage.CHECK_SERVICE)) {
            throw new WebSiteException(I18N.tr("TargoBankWebSite.webSiteServiceNotReachable"), //$NON-NLS-1$
                    WebSiteExceptionTypes.SERVICE_UNAVAILABLE);
        } else if (!body.getText().contains("Online-Postfach")) { //$NON-NLS-1$
            throw new WebSiteException(I18N.tr("TargoBankAccountStatementsPage.cannotOpenStatmentPage"), //$NON-NLS-1$
                    WebSiteExceptionTypes.SERVICE_UNAVAILABLE);
        }

        Logger.debug(this.driver.findElement(By.tagName("body")).getText()); //$NON-NLS-1$

        // get card account number and remove spaces
        final String cardAccountTrim = this.cardAccount.replaceAll(" ", StringUtils.EMPTY); //$NON-NLS-1$

        try {

            final WebElement selIndex1 = this.driver.findElement(By.name("SelIndex1")); //$NON-NLS-1$
            Logger.debug(selIndex1.getAttribute("outerHTML")); //$NON-NLS-1$

            final Select selection = new Select(selIndex1);

            WebElement optionForCard = null;

            final List<WebElement> options = selection.getOptions();
            for (final WebElement option : options) {
                Logger.debug(option.getText());
                if (option.getText().contains(cardAccountTrim)) {
                    optionForCard = option;
                    break;
                }
            }
            // Card account found? If not throw exception
            if (optionForCard == null) {
                Logger.error(I18N.tr("TargoBankAccountStatementsPage.cardAccountNotFound")); //$NON-NLS-1$ //NOPMD
                throw new WebSiteException(I18N.tr("TargoBankAccountStatementsPage.cardAccountNotFound"), //$NON-NLS-1$
                        WebSiteExceptionTypes.KONTO_NOT_FOUND);
            }

            // select card account
            selection.selectByValue(optionForCard.getAttribute("value")); //$NON-NLS-1$

            // get and press submit button
            final WebElement submitBtn = this.driver.findElement(By.id("btnRechercher")); //$NON-NLS-1$

            submitBtn.click();

        } catch (final NoSuchElementException e) {
            Logger.error(I18N.tr("TargoBankAccountStatementsPage.AccountsDropdownNotFound"), e); //$NON-NLS-1$
            throw new WebSiteException(I18N.tr("TargoBankAccountStatementsPage.AccountsDropdownNotFound"), //$NON-NLS-1$
                    WebSiteExceptionTypes.PAGE_NOT_FOUND, e);

        }
    }

    /**
     * Lde alle Umstze aus dem im Anchor angegebenen PDF
     *
     * @param anchor
     *            zu dem PDF.
     * @return Liste mit allen gefundenen Umstzen. Kann leer sein.
     */
    private List<Umsatz> loadTransactionsFromPDF(final WebElement link) {

        List<Umsatz> transactions = new ArrayList<>();

        try {

            // get the expected file name
            final String href = link.getAttribute(WebUtils.HTML_TAG_HREF);
            final URI uri = new URI(href);
            final Map<String, String> params = WebUtils.getQueryMap(uri);
            String expectedFileName = params.get("NOM"); //$NON-NLS-1$
            expectedFileName = expectedFileName.replace('+', ' ') + ".pdf"; //$NON-NLS-1$
            // get PDF as byte array
            final byte[] data = WebUtils.getSeleniumPdfAsByteArray(this.driver, link, expectedFileName);

            if (data == null) {
                Logger.error(I18N.tr("TargoBankAccountStatementsPage.loadPdfFailed", //$NON-NLS-1$
                        link.getText()));
            } else {

                Logger.debug("Lade PDF"); //$NON-NLS-1$
                final String cardAccountTrim = this.cardAccount.replaceAll(" ", StringUtils.EMPTY); //$NON-NLS-1$
                Logger.debug(I18N.tr("AbstractWebSite.extractTransactions", cardAccountTrim)); //$NON-NLS-1$
                final AbstractPdfExtractor extractor = new TargoBankPdfExtract(data, cardAccountTrim);
                final List<PdfBuchung> buchungenPDF = extractor.getExtract();

                // Ressourcen schlieen
                extractor.close();

                transactions = TransactionSaver.getTransactions(buchungenPDF);

            }
        } catch (IOException | InvalidCardNumber | URISyntaxException e) {
            Logger.error(I18N.tr("TargoBankAccountStatementsPage.readPdfContentFailed", e.getMessage()), e); //$NON-NLS-1$
        }

        return transactions;
    }

}
