/*******************************************************************************
 * Copyright 2015 - 2020 fmCreation
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald diese von der
 * Europischen Kommission genehmigt wurden - Folgeversionen der EUPL
 * ("Lizenz"); Sie drfen dieses Werk ausschlielich gem dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 *   https://progs.frankmuenster.de/eupl_v1-2_de
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften gefordert oder in
 * schriftlicher Form vereinbart, wird die unter der Lizenz verbreitete Software
 * "so wie sie ist", OHNE JEGLICHE GEWHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrcklich oder stillschweigend - verbreitet. Die sprachspezifischen
 * Genehmigungen und Beschrnkungen unter der Lizenz sind dem Lizenztext zu
 * entnehmen.
 ******************************************************************************/

package de.frankmuenster.mahoe.banks.targobank.web; // NOPMD

import java.rmi.RemoteException;
import java.text.SimpleDateFormat;
import java.util.Calendar;
import java.util.Date;
import java.util.List;
import java.util.Locale;

import org.apache.commons.lang.StringUtils;
import org.openqa.selenium.By;
import org.openqa.selenium.NoSuchSessionException;
import org.openqa.selenium.SessionNotCreatedException;
import org.openqa.selenium.WebElement;
import org.openqa.selenium.support.ui.ExpectedConditions;
import org.openqa.selenium.support.ui.WebDriverWait;

import de.frankmuenster.mahoe.MahoeSynchronizeBackend;
import de.frankmuenster.mahoe.banks.targobank.TargoBank;
import de.frankmuenster.mahoe.utils.I18N;
import de.frankmuenster.mahoe.utils.SecureCharSequence;
import de.frankmuenster.mahoe.utils.SeleniumCookie;
import de.frankmuenster.mahoe.utils.Utils;
import de.frankmuenster.mahoe.utils.WebUtils;
import de.frankmuenster.mahoe.web.AbstractSeleniumWebSite;
import de.frankmuenster.mahoe.web.AbstractWebSite;
import de.frankmuenster.mahoe.web.BankAccountStatementPage;
import de.frankmuenster.mahoe.web.MahoeKontoauszugPDF;
import de.frankmuenster.mahoe.web.WebSiteException;
import de.frankmuenster.mahoe.web.WebSiteExceptionTypes;
import de.willuhn.jameica.hbci.rmi.AuslandsUeberweisung;
import de.willuhn.jameica.hbci.rmi.Konto;
import de.willuhn.jameica.hbci.rmi.Umsatz;
import de.willuhn.logging.Logger;
import de.willuhn.util.ApplicationException;

/**
 * Stellt das Webseiten-Interface dar.
 *
 * @author Frank Mnster
 */
public class TargoBankWebSite extends AbstractSeleniumWebSite<TargoBankWebAccountInformation, TargoBankMessages> {

    /**
     * Seite mit den PDF-Abrechnungen nicht gefunden
     */
    protected static final String PDFSITE_NOT_FOUND = "Abrechnungen konnten nicht ge\u00F6ffnet werden"; //$NON-NLS-1$

    /**
     * Unsere BceeWebSite instance
     */
    private static TargoBankWebSite webSite;

    /**
     * Das Konto
     */
    private final Konto konto;

    /**
     * Das Login/die Kundenkennung
     */
    private final String kundenkennung;

    /**
     * Liefert die Instance der WebSeite
     *
     * @param konto
     * @return die WebSeite
     * @throws WebSiteException
     * @throws ApplicationException
     */
    public static TargoBankWebSite getInstance(final Konto konto) throws WebSiteException, ApplicationException {

        String cardAccount;
        String cardNumber;
        String cardCompany;
        String loginName;
        try {
            loginName = konto.getKundennummer();
            if (StringUtils.isEmpty(loginName)) {
                throw new ApplicationException(I18N.tr("TargoBankWebSite.loginMissing")); //$NON-NLS-1$
            }
            cardAccount = konto.getMeta(MahoeSynchronizeBackend.CARD_ACCOUNT, null);
            if (StringUtils.isEmpty(cardAccount)) {
                throw new ApplicationException(I18N.tr("TargoBankWebSite.cardAccountMissing")); //$NON-NLS-1$
            }
            cardNumber = konto.getMeta(MahoeSynchronizeBackend.CARD_NUMBER, null);
            if (StringUtils.isEmpty(cardNumber)) {
                throw new ApplicationException(I18N.tr("TargoBankWebSite.cardNumberMissing")); //$NON-NLS-1$
            }
            cardNumber = cardNumber.replaceAll("\\s+", StringUtils.EMPTY); //$NON-NLS-1$
            cardCompany = konto.getMeta(MahoeSynchronizeBackend.CARD_COMPANY, null);
            if (StringUtils.isEmpty(cardCompany)) {
                throw new ApplicationException(I18N.tr("TargoBankWebSite.cardCompanyMissing")); //$NON-NLS-1$
            }
            if (TargoBankWebSite.webSite == null) { // NOPMD
                TargoBankWebSite.webSite = new TargoBankWebSite(konto, loginName, cardCompany, cardAccount, cardNumber);
            } else {
                // Ist das Konto dasselbe
                if (TargoBankWebSite.webSite.konto.equals(konto)) {
                    // Aktualisiere Startdatum der Umsatzabfrage
                    final Date lastTransacDate = Utils.getLastUmsatzDate(konto,
                            TargoBankWebSite.webSite.getStartdate());
                    TargoBankWebSite.webSite.accountInfo.setLastTransactionDate(lastTransacDate);
                } else {
                    TargoBankWebSite.webSite.logoff();
                    TargoBankWebSite.webSite = new TargoBankWebSite(konto, loginName, cardCompany, cardAccount,
                            cardNumber);
                }
            }
        } catch (final RemoteException e) {
            final String msg = I18N.tr("TargoBankWebSite.readAccountInfoFailed"); //$NON-NLS-1$
            Logger.error(msg, e);
            throw new ApplicationException(msg, e);
        }
        return TargoBankWebSite.webSite;
    }

    /**
     * Erstellt ein neues Webseiten-Objekt zur Kontaktaufnahme mit der
     * TargoBank.
     *
     * @param konto
     * @param login
     * @param cardCompany
     * @param cardAccount
     * @param cardNumber
     * @throws WebSiteException
     */
    public TargoBankWebSite(final Konto konto, final String login, final String cardCompany, final String cardAccount,
            final String cardNumber) throws WebSiteException {

        super(konto);
        this.konto = konto;
        this.kundenkennung = login;
        this.accountInfo.setCardCompany(cardCompany);
        this.accountInfo.setCardNumber(cardNumber);
        this.accountInfo.setCardAccount(cardAccount);
    }

    /**
     * @see de.frankmuenster.mahoe.web.WebSite#createSepaTransfer(de.willuhn.jameica.hbci.rmi.AuslandsUeberweisung)
     */
    @Override
    public void createSepaTransfer(final AuslandsUeberweisung aUeberweisung)
            throws WebSiteException, ApplicationException {

        Logger.info(I18N.tr("AbstractWebSite.transferNotSupported")); //$NON-NLS-1$
    }

    /**
     * @see de.frankmuenster.mahoe.web.WebSite#getAccountInformation()
     */
    @Override
    public TargoBankWebAccountInformation getAccountInformation() throws WebSiteException, ApplicationException {

        final SimpleDateFormat sdf = new SimpleDateFormat("dd/MM/yyyy", Locale.getDefault()); //$NON-NLS-1$

        try {
            TargoBankWebSiteImpl.openAccount(this.webClient, this.accountInfo);

            final WebElement section = this.webClient.findElement(By.id("page")); //$NON-NLS-1$
            final List<WebElement> tables = section.findElements(By.tagName(WebUtils.HTML_TAG_TABLE));

            for (final WebElement table : tables) {

                final String summaryAttr = table.getAttribute("summary").trim(); //$NON-NLS-1$
                Logger.trace("Table summery: " + summaryAttr); //$NON-NLS-1$
                if ("Card Summary".equals(summaryAttr)) { //$NON-NLS-1$

                    String value;
                    // Saldo (R1/C2)
                    value = table.findElement(By.xpath("tbody/tr[1]/td[1]")).getText().trim() //$NON-NLS-1$
                            .replaceAll(AbstractWebSite.NON_BREAK_SPACE, StringUtils.EMPTY);
                    this.accountInfo.setSaldo(Utils.convertStringToBigDecimal(value));

                    // drawing limit (R2/C2)
                    value = table.findElement(By.xpath("tbody/tr[2]/td[1]")).getText().trim() //$NON-NLS-1$
                            .replaceAll(AbstractWebSite.NON_BREAK_SPACE, StringUtils.EMPTY);
                    this.accountInfo.setVerfuegungsrahmen(Utils.convertStringToDouble(value));

                    // available amount (R3/C2)
                    value = table.findElement(By.xpath("tbody/tr[3]/td[1]")).getText().trim() //$NON-NLS-1$
                            .replaceAll(AbstractWebSite.NON_BREAK_SPACE, StringUtils.EMPTY);
                    this.accountInfo.setSaldoAvailable(Utils.convertStringToBigDecimal(value));

                    // next due date (R3/C4)
                    value = table.findElement(By.xpath("tbody/tr[3]/td[2]")).getText().trim() //$NON-NLS-1$
                            .replaceAll(AbstractWebSite.NON_BREAK_SPACE, StringUtils.EMPTY);
                    final Date nextPayday = Utils.parseOrNull(sdf, value);
                    this.accountInfo.setNextPayDay(nextPayday);

                    Logger.info(I18N.tr("TargoBankWebSite.logSaldo") + this.accountInfo.getSaldo()); //$NON-NLS-1$
                    Logger.info(I18N.tr("TargoBankWebSite.logSalsoAvailable") + this.accountInfo.getSaldoAvailable()); //$NON-NLS-1$
                    Logger.info(I18N.tr("TargoBankWebSite.logVerfuegungsrahmen") //$NON-NLS-1$
                            + this.accountInfo.getVerfuegungsrahmen());
                    Logger.info(I18N.tr("TargoBankWebSite.logPayDay") + this.accountInfo.getSaldo()); //$NON-NLS-1$

                    break;

                }

            }

        } catch (final WebSiteException e) {
            Logger.error(I18N.tr("TargoBankWebSite.readAccountOverviewFailedLog", e.getMessage()), e); //$NON-NLS-1$
            throw new WebSiteException(I18N.tr("TargoBankWebSite.readAccountOverviewFailed"), // NOPMD //$NON-NLS-1$
                    WebSiteExceptionTypes.READ_ACCOUNT_DETAILS_FAILED, e);
        }
        return this.accountInfo;
    }

    @Override
    public List<MahoeKontoauszugPDF> getAccountSheets(final TargoBankWebAccountInformation accountInfo)
            throws WebSiteException {

        final TargoBankAccountStatementFinder accountSheets = new TargoBankAccountStatementFinder(this.webClient,
                accountInfo);
        accountSheets.findStatements();

        return accountSheets.getSheets();

    }

    /**
     * @see de.frankmuenster.mahoe.web.WebSite#getBookedTransactions()
     */
    @Override
    public List<Umsatz> getBookedTransactions() throws WebSiteException, ApplicationException {

        // open the account statements page
        BankAccountStatementPage page = new TargoBankAccountStatementsPage(this.webClient,
                this.accountInfo.getCardAccount());

        // two month back only to be save
        final Date from = new Date(this.accountInfo.getLastTransactionDate().getTime() - 60 * 24 * 60 * 60 * 1000);

        // get transactions from actual page
        final List<Umsatz> transactions = page.getTransactions(from);

        // check if we need to load from other pages as well
        while (page.accountLinksHasNextPage()) {
            Logger.debug("More pages available"); //$NON-NLS-1$
            page = page.getNextPage();
            transactions.addAll(page.getTransactions(from));
        }

        return transactions;
    }

    /**
     * @see de.frankmuenster.mahoe.web.WebSite#getPreBookedTransactions()
     */
    @Override
    public List<Umsatz> getPreBookedTransactions() throws WebSiteException, ApplicationException {

        final TargoBankWebSiteTransactionsImpl transImpl = new TargoBankWebSiteTransactionsImpl(this.webClient,
                this.accountInfo);
        return transImpl.getPreBookTransactions();
    }

    /**
     * @see de.frankmuenster.mahoe.web.WebSite#isStillLoggedIn()
     */
    @Override
    public boolean isStillLoggedIn() throws WebSiteException {

        boolean stillLoggedIn = false;

        // first check if driver available
        this.checkSeleniumDriverSet(this.webClient);

        try {
            stillLoggedIn = this.elementExits(By.id("menu_mein")); //$NON-NLS-1$
        } catch (final NoSuchSessionException | SessionNotCreatedException e) {
            this.getWebClient();
        }

        return stillLoggedIn;

    }

    /**
     * @see de.frankmuenster.mahoe.web.WebSite#login(char[], String...)
     */
    @Override
    public void login(final char[] password, final String... loginParams) throws WebSiteException {

        // first check if driver available
        this.checkSeleniumDriverSet(this.webClient);

        if (loginParams.length != 0) {
            throw new WebSiteException(I18N.tr("TargoBankWebSite.loginParameterNumberWrong"), //$NON-NLS-1$
                    WebSiteExceptionTypes.LOGIN_ERROR);
        }

        final WebDriverWait waitNormal = new WebDriverWait(this.webClient, 10);
        final WebDriverWait waitCaptcha = new WebDriverWait(this.webClient, 300);

        // Open Targobank loogin page
        this.webClient.get(TargoBank.URL_LOGIN);

        if (!this.webClient.findElements(By.id("acceptAllCookiesButton")).isEmpty()) { //$NON-NLS-1$
            final WebElement cookieWarning = this.webClient.findElement(By.id("acceptAllCookiesButton")); //$NON-NLS-1$
            cookieWarning.click();
        }

        // Restore cookies
        SeleniumCookie.read(this.webClient, TargoBank.COOKIE_FILE);

        final WebElement login = waitNormal.until(ExpectedConditions.elementToBeClickable(By.id("login-submit"))); //$NON-NLS-1$

        final WebElement user = this.webClient.findElement(By.name("_cm_user")); //$NON-NLS-1$
        user.sendKeys(this.kundenkennung); // $NON-NLS-1$

        final WebElement pwd = this.webClient.findElement(By.name("_cm_pwd")); //$NON-NLS-1$
        final SecureCharSequence securePW = new SecureCharSequence(password);
        pwd.sendKeys(securePW);
        securePW.wipe();

        // perform login
        login.click();

        // wait until ReCaptcha and TAN has been provided
        waitCaptcha.until(ExpectedConditions.presenceOfElementLocated(By.id("e_menuLevel1"))); //$NON-NLS-1$

        // do a final check, if login was successful
        final String check = this.webClient.getPageSource();
        if (!check.contains("Meine TARGOBANK")) { //$NON-NLS-1$
            throw new WebSiteException(I18N.tr("TargoBankWebSite.loginFailed"), WebSiteExceptionTypes.LOGIN_ERROR); //$NON-NLS-1$
        }
        if (check.contains("Die von Ihnen eingegebenen Daten sind falsch. Bitte versuchen Sie es erneut.")) { //$NON-NLS-1$
            Logger.warn(I18N.tr("TargoBankWebSite.loginDataWrong")); //$NON-NLS-1$
            throw new WebSiteException(I18N.tr("TargoBankWebSite.loginDataWrong"), WebSiteExceptionTypes.LOGIN_FAILED); //$NON-NLS-1$
        }

    }

    /**
     * @see de.frankmuenster.mahoe.web.WebSite#logoff()
     */
    @Override
    public void logoff() throws WebSiteException {

        // call log off URL
        this.webClient.get(TargoBank.URL_LOGOFF);

        // check if log off was successful
        final String title = this.webClient.getTitle();
        if (title.startsWith("Online-Banking Logout")) { //$NON-NLS-1$
            Logger.info(I18N.tr("TargoBankWebSite.logoffSuccessful")); //$NON-NLS-1$
        } else {
            throw new WebSiteException(I18N.tr("TargoBankWebSite.logoffFailed"), WebSiteExceptionTypes.LOGOFF_FAILED); //$NON-NLS-1$
        }

        // finally save the cookies
        saveCookies(TargoBank.COOKIE_FILE);
    }

    /**
     * @see de.frankmuenster.mahoe.web.AbstractWebSite#getStartdate()
     */
    @Override
    protected Calendar getStartdate() {

        // Setzte das startDatum auf Januar 2016
        final Calendar startCalendar = Calendar.getInstance();
        startCalendar.set(2016, 0, 1);
        return startCalendar;
    }

}
