/*******************************************************************************
 * Copyright 2015 - 2019 fmCreation
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald diese von der
 * Europischen Kommission genehmigt wurden - Folgeversionen der EUPL
 * ("Lizenz"); Sie drfen dieses Werk ausschlielich gem dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 *   https://progs.frankmuenster.de/eupl_v1-2_de
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften gefordert oder in
 * schriftlicher Form vereinbart, wird die unter der Lizenz verbreitete Software
 * "so wie sie ist", OHNE JEGLICHE GEWHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrcklich oder stillschweigend - verbreitet. Die sprachspezifischen
 * Genehmigungen und Beschrnkungen unter der Lizenz sind dem Lizenztext zu
 * entnehmen.
 ******************************************************************************/

package de.frankmuenster.mahoe.banks.targobank.web; // NOPMD

import java.rmi.RemoteException;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;
import java.util.Locale;

import org.apache.commons.lang.StringUtils;
import org.openqa.selenium.By;
import org.openqa.selenium.NoSuchElementException;
import org.openqa.selenium.WebDriver;
import org.openqa.selenium.WebElement;

import de.frankmuenster.mahoe.utils.I18N;
import de.frankmuenster.mahoe.utils.Utils;
import de.frankmuenster.mahoe.utils.WebUtils;
import de.frankmuenster.mahoe.web.AbstractWebSite;
import de.frankmuenster.mahoe.web.WebSiteException;
import de.willuhn.jameica.hbci.Settings;
import de.willuhn.jameica.hbci.rmi.HBCIDBService;
import de.willuhn.jameica.hbci.rmi.Umsatz;
import de.willuhn.logging.Logger;
import de.willuhn.util.ApplicationException;

/***
 * Statische Implementierungen der Umsatzfunktionen
 *
 * @author Frank Mnster
 */
public class TargoBankWebSiteTransactionsImpl {

    /**
     * Die Kontoinformationen
     */
    private final TargoBankWebAccountInformation accountInfo;

    /**
     * This is the driver instance.
     */
    protected WebDriver webclient;

    /**
     * Konstruktor
     *
     * @param webclient
     * @param accountInfo
     */
    public TargoBankWebSiteTransactionsImpl(final WebDriver webclient,
            final TargoBankWebAccountInformation accountInfo) {

        this.accountInfo = accountInfo;
        this.webclient = webclient;
    }

    /**
     * Liste die vorgemerkten Umstze ein.
     *
     * @return ein Liste mit den vorgemerkten Umstzen
     * @throws WebSiteException
     * @throws ApplicationException
     */
    protected List<Umsatz> getPreBookTransactions() throws WebSiteException, ApplicationException {

        List<Umsatz> umsaetze = new ArrayList<>();

        try {
            TargoBankWebSiteImpl.openAccount(this.webclient, this.accountInfo);

            final WebElement section = this.webclient.findElement(By.id("page")); //$NON-NLS-1$
            final List<WebElement> tables = section.findElements(By.tagName(WebUtils.HTML_TAG_TABLE));

            boolean isNewPage = true;
            while (isNewPage) {

                for (final WebElement table : tables) {

                    final String summaryAttr = table.getAttribute("summary").trim(); //$NON-NLS-1$
                    Logger.trace("Table summery: " + summaryAttr); //$NON-NLS-1$
                    if ("Activity Details".equals(summaryAttr)) { //$NON-NLS-1$
                        umsaetze = this.readWebSiteTable(table);
                    }
                    isNewPage = this.existsNextPage();
                }
            }

        } catch (final WebSiteException e) {
            Logger.error(I18N.tr("TargoBankWebSiteTransactionsImpl.readPreBookFailed", e.getMessage())); //$NON-NLS-1$
        }

        // Add saldo to transactions
        double zwischenSaldo = this.accountInfo.getSaldo().doubleValue();
        for (final Umsatz umsatz : umsaetze) {
            try {
                umsatz.setSaldo(zwischenSaldo);
                zwischenSaldo = Utils.roundNumber(zwischenSaldo - umsatz.getBetrag(), 2);
            } catch (final RemoteException e) {
                Logger.warn(I18N.tr("TargoBankWebSiteTransactionsImpl.calculatingSaldoFailed", e.getMessage())); //$NON-NLS-1$
            }
        }

        return umsaetze;

    }

    /**
     * Check if a next button exits and load the page
     *
     * @return <i>true</i> if a next page exists and has been loaded,
     *         <i>false</i> else.
     */
    private boolean existsNextPage() {

        // check for next page
        // TODO: Test auf weitere Tabellenseiten
        // Konnte noch nicht erfolgen, da ich zu wenig
        // Umsaetze habe.

        return false;
    }

    /**
     * Liest die vorgemerkten Umstze.
     *
     * @param table
     *            die Table mit dem Umstzen
     * @return eine Liste mit den Umstzen
     */
    private List<Umsatz> readWebSiteTable(final WebElement table) {

        final List<Umsatz> umsaetze = new ArrayList<>();
        final SimpleDateFormat sdf = new SimpleDateFormat("dd/MM/yyyy", Locale.getDefault()); //$NON-NLS-1$

        final List<WebElement> bodies = table.findElements(By.tagName(WebUtils.HTML_TAG_TBODY));
        for (final WebElement body : bodies) {
            final List<WebElement> rows = body.findElements(By.tagName(WebUtils.HTML_TAG_TR));
            for (final WebElement row : rows) {
                Logger.debug(row.getText());
                try {

                    if (row.findElements(By.xpath(WebUtils.HTML_TAG_TD1)).isEmpty()) {
                        // Header line --> skip
                        continue;
                    }
                    // Create new transaction
                    final HBCIDBService dbService = Settings.getDBService();
                    final Umsatz umsatz = dbService.createObject(Umsatz.class, null);

                    // date
                    final String datumStr = row.findElement(By.xpath(WebUtils.HTML_TAG_TD1)).getText().trim()
                            .replaceAll(AbstractWebSite.NON_BREAK_SPACE, StringUtils.EMPTY);
                    Logger.debug(I18N.tr("TargoBankWebSiteTransactionsImpl.transactionDate", datumStr)); //$NON-NLS-1$

                    // Zweck, besteht aus Details, Stadt und Land
                    final String details = row.findElement(By.xpath("td[2]")).getText().trim() //$NON-NLS-1$
                            .replaceAll(AbstractWebSite.NON_BREAK_SPACE, StringUtils.EMPTY);
                    final String stadt = row.findElement(By.xpath("td[3]")).getText().trim() //$NON-NLS-1$
                            .replaceAll(AbstractWebSite.NON_BREAK_SPACE, StringUtils.EMPTY);
                    final String land = row.findElement(By.xpath("td[4]")).getText().trim() //$NON-NLS-1$
                            .replaceAll(AbstractWebSite.NON_BREAK_SPACE, StringUtils.EMPTY);

                    final String zweck = details + " " + stadt + " " //$NON-NLS-1$ //$NON-NLS-2$
                            + land;
                    Utils.loadUmsatz(umsatz, zweck);

                    // Betrag
                    final String betrag = row.findElement(By.xpath("td[5]")).getText().trim() //$NON-NLS-1$
                            .replaceAll(AbstractWebSite.NON_BREAK_SPACE, StringUtils.EMPTY);
                    umsatz.setBetrag(Utils.convertStringToDouble(betrag));

                    // Fremdwhungsbetrag
                    final String fcValue = row.findElement(By.xpath("td[6]")).getText().trim() //$NON-NLS-1$
                            .replaceAll(AbstractWebSite.NON_BREAK_SPACE, StringUtils.EMPTY);
                    if (StringUtils.isNotEmpty(fcValue)) {
                        umsatz.setKommentar(I18N.tr("TargoBankWebSiteTransactionsImpl.foreignCurrencySaldo", fcValue)); //$NON-NLS-1$
                    }

                    // Buchungsdatum
                    final Date datum = Utils.parseOrNull(sdf, datumStr);
                    umsatz.setDatum(datum);
                    umsatz.setValuta(datum);
                    umsatz.setFlags(Umsatz.FLAG_NOTBOOKED);
                    umsaetze.add(umsatz);
                } catch (final RemoteException | NoSuchElementException re) {
                    Logger.error(I18N.tr("TargoBankWebSiteTransactionsImpl.createUmsatzError", re.getMessage()), re); //$NON-NLS-1$
                }
            } // end for weRows
        } // end for weBodies

        return umsaetze;
    }
}
