package de.frankmuenster.mahoe.gui.utils;

import java.util.HashMap;
import java.util.Locale;
import java.util.Map;

import org.apache.commons.lang.StringUtils;
import org.eclipse.swt.SWT;
import org.eclipse.swt.graphics.Device;
import org.eclipse.swt.graphics.Font;
import org.eclipse.swt.graphics.FontData;
import org.eclipse.swt.widgets.Display;
import org.junit.Assert;

/**
 * {@link SWT} font related utility methods.
 */
@SuppressWarnings({ "PMD", "nls" })
public class SWTFontUtils {

    /**
     * Font Consolas|normal|10
     */
    private static final String CONSOLAS_NORMAL_10 = "Consolas|normal|10";
    /**
     * Font Lucida Console|normal|9
     */
    private static final String LUCIDA_CONSOLE_NORMAL_9 = "Lucida Console|normal|9";
    /**
     * Font Courier|normal|10
     */
    private static final String COURIER_NORMAL_10 = "Courier|normal|10";
    /**
     * Font Courier New|normal|10
     */
    private static final String COURIER_NEW_NORMAL_10 = "Courier New|normal|10";
    /**
     * Font adobe-courier|normal|12
     */
    private static final String ADOBE_COURIER_NORMAL_12 = "adobe-courier|normal|12";

    /** Cache: mapping from SWT devices to their monospaced fonts. */
    private static final Map<Device, Font> MONOSPACED_FONTS = new HashMap<>();

    /**
     * Returns the monospaced font for the current display. The font will
     * automatically be disposed once the display is disposed.
     * <p>
     * This method is thread safe.
     * </p>
     *
     * @return The monospaced font for the current display.
     * @throws IllegalStateException
     *             If the method is not invoked from a SWT UI thread.
     * @throws AssertionError
     * @throws IllegalArgumentException
     */
    public static Font getMonospacedFont() throws AssertionError {

	synchronized (SWTFontUtils.MONOSPACED_FONTS) {
	    // Get current display.
	    final Display display = Display.getCurrent();
	    if (display == null) {
		final String msg = "Must be invoked for a SWT UI thread.";
		throw new IllegalStateException(msg);
	    }

	    // Forward request.
	    return SWTFontUtils.getMonospacedFont(display);
	}
    }

    /**
     * Creates a monospaced font for the given display. The font will
     * automatically be disposed once the display is disposed.
     * <p>
     * This method is thread safe.
     * </p>
     *
     * @param display
     *            The display for which to create a monospaced font.
     * @return A monospaced font for the given display.
     * @throws AssertionError
     * @throws IllegalArgumentException
     */
    public static Font getMonospacedFont(final Display display) throws AssertionError {

	synchronized (SWTFontUtils.MONOSPACED_FONTS) {
	    // Based on class 'org.eclipse.jface.resource.FontRegistry' and
	    // resources 'org.eclipse.jface.resource/jfacefonts*.properties'
	    // from 'org.eclipse.jface' plug-in (version 3.9.1).

	    // Use cache if possible.
	    final Font cachedFont = SWTFontUtils.MONOSPACED_FONTS.get(display);
	    if (cachedFont != null) {
		return cachedFont;
	    }

	    // Get operating system and windowing system names.
	    String os = System.getProperty("os.name");
	    String ws = SWT.getPlatform();
	    os = StringUtils.deleteWhitespace(os).toLowerCase(Locale.US);
	    ws = StringUtils.deleteWhitespace(ws).toLowerCase(Locale.US);

	    // Get names to check, in order from specific to generic.
	    final String[] names = { os + "_" + ws, os, "" };

	    // Get font data texts for platform.
	    String[] fontDataTxts = null;
	    for (final String name : names) {

		switch (name) {

		case "aix":
		    fontDataTxts = new String[] { SWTFontUtils.ADOBE_COURIER_NORMAL_12 };
		    break;

		case "hp-ux":
		    fontDataTxts = new String[] { "adobe-courier|normal|14" };
		    break;

		case "linux_gtk":
		    fontDataTxts = new String[] { "Monospace|normal|10" };
		    break;

		case "linux":
		    fontDataTxts = new String[] { SWTFontUtils.ADOBE_COURIER_NORMAL_12 };
		    break;

		case "macosx":
		    fontDataTxts = new String[] { "Monaco|normal|11", "Courier|normal|12", "Courier New|normal|12" };
		    break;

		case "sunos":
		case "solaris":
		    fontDataTxts = new String[] { SWTFontUtils.ADOBE_COURIER_NORMAL_12 };
		    break;

		case "windows98":
		case "windowsnt":
		case "windows2000":
		case "windowsxp":
		    fontDataTxts = new String[] { SWTFontUtils.COURIER_NEW_NORMAL_10, SWTFontUtils.COURIER_NORMAL_10,
			    SWTFontUtils.LUCIDA_CONSOLE_NORMAL_9 };
		    break;

		case "windowsvista":
		    fontDataTxts = new String[] { SWTFontUtils.CONSOLAS_NORMAL_10, SWTFontUtils.COURIER_NEW_NORMAL_10 };
		    break;

		case "windows7":
		    fontDataTxts = new String[] { SWTFontUtils.CONSOLAS_NORMAL_10, SWTFontUtils.COURIER_NEW_NORMAL_10 };
		    break;

		case "windows8":
		    fontDataTxts = new String[] { SWTFontUtils.CONSOLAS_NORMAL_10, SWTFontUtils.COURIER_NEW_NORMAL_10 };
		    break;

		default:
		    fontDataTxts = new String[] { SWTFontUtils.COURIER_NEW_NORMAL_10, SWTFontUtils.COURIER_NORMAL_10,
			    "b&h-lucidabright|normal|9" };
		    break;
		}
	    }

	    if (fontDataTxts == null) {
		// Can't happen, but silences a warning.
		throw new AssertionError();
	    }

	    // Convert texts to font data.
	    final FontData[] fontDatas = new FontData[fontDataTxts.length];
	    for (int i = 0; i < fontDatas.length; i++) {
		// Find splitters.
		final String txt = fontDataTxts[i];
		final int bar2 = txt.lastIndexOf('|');
		Assert.assertTrue(bar2 != -1);
		final int bar1 = txt.lastIndexOf('|', bar2 - 1);
		Assert.assertTrue(bar1 != -1);

		// Get font name.
		final String name = txt.substring(0, bar1);
		Assert.assertTrue(name.length() > 0);

		// Get font style.
		final String[] styles = txt.substring(bar1 + 1, bar2).split(",");
		int style = 0;
		for (final String s : styles) {
		    switch (s) {
		    case "normal":
			style |= SWT.NORMAL;
			break;
		    case "bold":
			style |= SWT.BOLD;
			break;
		    case "italic":
			style |= SWT.ITALIC;
			break;
		    default:
			throw new IllegalArgumentException("Invalid style: " + s);
		    }
		}

		// Get font height.
		final int height = Integer.parseInt(txt.substring(bar2 + 1));

		// Create and add font date.
		fontDatas[i] = new FontData(name, height, style);
	    }

	    // Create font.
	    final Font font = new Font(display, fontDatas);

	    // Register dispose callback, to dispose of the font once the
	    // display is disposed.
	    display.disposeExec(() -> {
		synchronized (SWTFontUtils.MONOSPACED_FONTS) {
		    SWTFontUtils.MONOSPACED_FONTS.remove(display);
		    font.dispose();
		}
	    });

	    // Add to cache.
	    SWTFontUtils.MONOSPACED_FONTS.put(display, font);

	    // Return the new font.
	    return font;
	}

    }

    /** Constructor for the {@link SWTFontUtils} class. */
    private SWTFontUtils() {
	// Static class.
    }
}
