/*******************************************************************************
 * Copyright 2015 - 2020 fmCreation
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald diese von der
 * Europischen Kommission genehmigt wurden - Folgeversionen der EUPL
 * ("Lizenz"); Sie drfen dieses Werk ausschlielich gem dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 *   https://progs.frankmuenster.de/eupl_v1-2_de
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften gefordert oder in
 * schriftlicher Form vereinbart, wird die unter der Lizenz verbreitete Software
 * "so wie sie ist", OHNE JEGLICHE GEWHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrcklich oder stillschweigend - verbreitet. Die sprachspezifischen
 * Genehmigungen und Beschrnkungen unter der Lizenz sind dem Lizenztext zu
 * entnehmen.
 ******************************************************************************/

package de.frankmuenster.mahoe.web;

import java.io.File;
import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.Set;
import java.util.concurrent.TimeUnit;

import org.apache.commons.io.FileUtils;
import org.apache.commons.lang.StringUtils;
import org.openqa.selenium.By;
import org.openqa.selenium.Cookie;
import org.openqa.selenium.NoSuchSessionException;
import org.openqa.selenium.SessionNotCreatedException;
import org.openqa.selenium.WebDriver;
import org.openqa.selenium.WebDriverException;
import org.openqa.selenium.firefox.FirefoxDriver;
import org.openqa.selenium.firefox.FirefoxOptions;
import org.openqa.selenium.firefox.FirefoxProfile;

import de.frankmuenster.mahoe.MahoeMain;
import de.frankmuenster.mahoe.Settings;
import de.frankmuenster.mahoe.messages.AbstractBankMessages;
import de.frankmuenster.mahoe.utils.I18N;
import de.frankmuenster.mahoe.utils.SeleniumCookie;
import de.willuhn.jameica.hbci.rmi.Konto;
import de.willuhn.jameica.messaging.StatusBarMessage;
import de.willuhn.jameica.system.Application;
import de.willuhn.logging.Logger;
import de.willuhn.util.ApplicationException;

/**
 * Abstract class for Selenium based WebSite to provide the driver and a temp
 * download folder.
 *
 * @author Frank Mnster
 * @param <I>
 *            the concrete WebAccountInformation
 * @param <M>
 *            the concrete AbstractBankMessages
 * @since 2.0
 */
public abstract class AbstractSeleniumWebSite<I extends WebAccountInformation,
        M extends AbstractBankMessages<WebDriver>> extends AbstractWebSite<WebDriver, I, M> {

    /**
     * Mime type for downloads
     */
    private static final String MIME_TYPE_PDF = "application/pdf"; //$NON-NLS-1$

    /**
     * Lock to synchronise access to tmpfolder method
     */
    private static final Object LOCK = new Object() {
    };

    /**
     * A static webdriver across all implementations
     */
    private static WebDriver driver;

    /**
     * Temporary folder to download the account sheets for further processing
     */
    private static Path tmpDownloadFld;

    /**
     * Returns the temporary folder to use for downloading account sheets.
     * Create a new one per session.
     *
     * @return the tmp folder to use during this session
     * @throws ApplicationException
     *             if the tmp folder cannot be created
     */
    public static Path getDownloadsDirectory() throws ApplicationException {

        synchronized (AbstractSeleniumWebSite.LOCK) {

            if (AbstractSeleniumWebSite.tmpDownloadFld == null) {

                try {
                    AbstractSeleniumWebSite.tmpDownloadFld = Files.createTempDirectory("selleniumdownloads_"); //$NON-NLS-1$
                } catch (final IOException ex) {
                    throw new ApplicationException(I18N.tr("AbstractSeleniumWebSite.failedToCreateTmpFolder"), ex); //$NON-NLS-1$
                }
            } else {
                // check if folder still exists
                if (!AbstractSeleniumWebSite.tmpDownloadFld.toFile().exists()) {
                    try {
                        Files.createDirectory(AbstractSeleniumWebSite.tmpDownloadFld);
                    } catch (final IOException e) {
                        Logger.error(e.getLocalizedMessage(), e);
                    }
                }
            }
        }
        return AbstractSeleniumWebSite.tmpDownloadFld;
    }

    /**
     * Clean and remove temp-folder and shutdown the driver
     */
    public static void shutDown() {

        // End driver
        if (AbstractSeleniumWebSite.driver != null) {
            AbstractSeleniumWebSite.driver.quit();
        }

        // delete tmp folder
        try {
            final Path dir = AbstractSeleniumWebSite.getDownloadsDirectory();
            FileUtils.deleteDirectory(dir.toFile());
        } catch (final IOException | ApplicationException e) {
            Logger.error(I18N.tr("AbstractSeleniumWebSite.failedToDeleteTmpFolder"), e); //$NON-NLS-1$
        }

    }

    /**
     * Standardkonstruktor
     *
     * @param konto
     * @throws WebSiteException
     */
    protected AbstractSeleniumWebSite(final Konto konto) throws WebSiteException {

        super(konto);
    }

    /**
     * Save the cookies to be used next time
     */
    public void saveCookies(final String backupFileName) {

        // save cookies
        final Set<Cookie> cok = AbstractSeleniumWebSite.driver.manage().getCookies();
        Logger.debug("Cookies: " + cok.toString()); //$NON-NLS-1$

        try {
            SeleniumCookie.write(AbstractSeleniumWebSite.driver, backupFileName);
        } catch (final ApplicationException e) {

            // Display as warning
            final StatusBarMessage msg = new StatusBarMessage(
                    "Browser-Cookies konnten nicht gespeichert werden. Evtl. mussen Logindaten fter eingegeben werden.", //$NON-NLS-1$
                    StatusBarMessage.TYPE_INFO);
            Application.getMessagingFactory().sendMessage(msg);
        }

    }

    /**
     * Checks if the driver is set and active. if not a status bar message is
     * send that driver is not configured.
     *
     * @param driver
     *            the driver to check
     * @throws WebSiteException
     *             when the driver is not set
     */
    protected void checkSeleniumDriverSet(final WebDriver driver) throws WebSiteException {

        // the driver set
        if (driver == null) {

            // create a new driver
            this.getWebClient();

        }

        // do we have a session
        try {
            this.webClient.getTitle();
            // browser is open
        } catch (final NoSuchSessionException | SessionNotCreatedException e) {
            // start a new one
            this.getWebClient();
        }

    }

    /**
     * checks if an element identified by {@link By} exits
     *
     * @param byId
     *            the element to search for
     * @return <i>truer</i> in case the element exists, <i>false</i> else
     */
    protected boolean elementExits(final By byId) {

        this.webClient.manage().timeouts().implicitlyWait(0, TimeUnit.MILLISECONDS);
        final boolean exists = this.webClient.findElements(byId).size() != 0;
        this.webClient.manage().timeouts().implicitlyWait(3, TimeUnit.SECONDS);
        return exists;

    }

    /**
     * Here we setup the Selenium driver as replacement for the HTMLUnit client.
     */
    @Override
    protected synchronized WebDriver getWebClient() { // NOPMD

        if (AbstractSeleniumWebSite.driver == null && this.webClient != null) {
            AbstractSeleniumWebSite.driver = this.webClient;
        }

        if (AbstractSeleniumWebSite.driver != null) {
            // check for noSession
            try {
                AbstractSeleniumWebSite.driver.getPageSource();

                // set webClient only to be sure
                this.webClient = AbstractSeleniumWebSite.driver;
            } catch (final WebDriverException e) {
                AbstractSeleniumWebSite.driver = null; // NOPMD on 13.11.19,
                                                       // 19:15
                this.webClient = null; // NOPMD on 13.11.19, 19:15
            }
        }

        // Only if the the driver is not already created
        if (this.webClient == null && AbstractSeleniumWebSite.driver == null) {

            // create a new driver
            createNewDriver();
            // assign driver to webclient
            this.webClient = AbstractSeleniumWebSite.driver;

        } else {
            // check temp folder
            if (!AbstractSeleniumWebSite.tmpDownloadFld.toFile().exists()) {
                try {
                    Files.createDirectory(AbstractSeleniumWebSite.tmpDownloadFld);
                } catch (final IOException e) {
                    Logger.error(e.getLocalizedMessage(), e);
                }
            }
        }

        return AbstractSeleniumWebSite.driver;

    }

    /**
     * Creates a new Selenium web driver. </br>
     * It also creates a new tmpFolder if needed. Furthermore it displays an
     * error message in the GUI in case the driver could not be created.
     */
    private void createNewDriver() {
        String tmpdir = StringUtils.EMPTY;
        try {
            tmpdir = AbstractSeleniumWebSite.getDownloadsDirectory().toString();
        } catch (final ApplicationException e) {
            Logger.error(e.getLocalizedMessage(), e);
        }

        // path to plugin profile
        final String path = Application.getPluginLoader().getPlugin(MahoeMain.class).getManifest().getPluginDir();
        final File sourceProfile = new File(path, "ffProfile"); //$NON-NLS-1$

        // create Profile from default one
        final FirefoxProfile fxProfile = new FirefoxProfile(sourceProfile);

        // Some settings for auto download
        fxProfile.setPreference("browser.download.folderList", 2); //$NON-NLS-1$
        fxProfile.setPreference("browser.download.manager.showWhenStarting", false); //$NON-NLS-1$
        fxProfile.setPreference("browser.download.dir", tmpdir); //$NON-NLS-1$
        fxProfile.setPreference("browser.download.useDownloadDir", true); //$NON-NLS-1$
        fxProfile.setPreference("browser.helperApps.neverAsk.saveToDisk", AbstractSeleniumWebSite.MIME_TYPE_PDF); //$NON-NLS-1$
        fxProfile.setPreference("browser.helperApps.neverAsk.openFile", AbstractSeleniumWebSite.MIME_TYPE_PDF); //$NON-NLS-1$
        fxProfile.setPreference("browser.helperApps.alwaysAsk.force", false); //$NON-NLS-1$
        fxProfile.setPreference("pdfjs.disabled", true); //$NON-NLS-1$
        fxProfile.setPreference("pdfjs.enabledCache.state", false); //$NON-NLS-1$

        // New Firefox option
        final FirefoxOptions options = new FirefoxOptions();
        // Add profile
        options.setProfile(fxProfile);

        // Set the path to the Selenium driver
        final String selDriver = Settings.getSeleniumDriver();
        final Path driverPath = Paths.get(selDriver);
        if (StringUtils.isNotEmpty(selDriver) && driverPath.toFile().exists()) {

            // Instantiate a new Firefox driver
            System.setProperty("webdriver.gecko.driver", selDriver); //$NON-NLS-1$
            synchronized (AbstractSeleniumWebSite.LOCK) {
                AbstractSeleniumWebSite.driver = new FirefoxDriver(options);
            }

        } else {
            Logger.warn(I18N.tr("AbstractSeleniumWebSite.selDriverNotSet")); //$NON-NLS-1$ )
            final StatusBarMessage msg = new StatusBarMessage(I18N.tr("AbstractSeleniumWebSite.selDriverNotSetMsg"), //$NON-NLS-1$
                    StatusBarMessage.TYPE_ERROR);
            Application.getMessagingFactory().sendMessage(msg);
        }
    }

}
